import hashlib
import json
import os
import random
import string
from typing import Union, Dict

from PIL import Image as PIL_IMAGE
from assertpy import assert_that

from deepdriver import logger
from deepdriver.sdk.data_types.boundingBoxes import BoundingBoxes
from deepdriver.sdk.data_types.media import LOG_TYPE_IMAGE, Media
from deepdriver.sdk.data_types.run import Run
from deepdriver.sdk.interface import interface


class Image(Media):

    def __init__(self, data: Union[str, PIL_IMAGE.Image], caption: str = None,
                 boxes: Union[Dict[str, BoundingBoxes], Dict[str, dict]] = None) -> None:
        super().__init__(log_type=LOG_TYPE_IMAGE)

        assert_that(data).is_not_none()
        if isinstance(data, str):
            # str인 경우 local_path로 지정하고 data 에는 해당 경로 파일을 로드한 PIL.Image 객체를 지정함
            self.local_path = data
            self.data = PIL_IMAGE.open(self.local_path)
            self.height = self.data.size[1]
            self.width = self.data.size[0]
            self.format = self.data.format
        elif isinstance(data, PIL_IMAGE.Image):
            self.data = data
            # data가 PIL.Image 객체인 경우 7자리의 random id값을 생성한후 {random_id}.{file_format}으로 저장
            random_id = "".join(random.choice(string.digits) for _ in range(7))
            self.local_path = f"{random_id}.{data.format}"
            logger.debug(f"PIL Image save to local:local_path:{self.local_path}")
            data.save(self.local_path)
            self.height = self.data.size[1]
            self.width = self.data.size[0]
            self.format = self.data.format
        else:
            raise Exception(f"unknown data type {type(data)}")

        # caption, boxes 처리
        self.caption :str = caption
        self.boxes :Dict = None

        if boxes:
            if isinstance(boxes, BoundingBoxes):
                self.boxes = boxes
            elif isinstance(boxes, dict):
                bb = BoundingBoxes(
                    key=list(boxes.keys())[0],
                    box_data=list(boxes.values())[0]['box_data'],
                    class_labels=list(boxes.values())[0]['class_labels'])
                self.boxes = bb



    def to_json(self, key_name: str) -> str:
        assert_that(key_name).is_not_none()

        return json.dumps({
            "log_type": self.log_type,
            "hash": self.get_digest(),
            "size": self.get_size(),
            "path": self.get_path(key_name),
            "height": self.height,
            "width": self.width,
            "format": self.format,
            "caption": self.caption if self.caption else "",
            "boxes": self.boxes.to_dict() if self.boxes else "",

        })

    def get_size(self) -> int:
        return os.stat(self.local_path).st_size

    def get_digest(self) -> str:
        with open(self.local_path, "rb") as f:
            digest = hashlib.md5(f.read()).hexdigest()
        return digest

    # json으로 구성된 메타데이터 전송
    def upload_file(self, run: Run, key_name: str) -> None:
        # 저장한 파일을 Interface.py의 upoad_file을 호출하여 전송
        digest = self.get_digest()
        root_path = self.get_root_path(run.run_id)
        path = self.get_path(key_name)
        logger.debug(f"file upload[image] : local_path=[{self.local_path}], root_path=[{root_path}], path=[{path}]")
        interface.upload_file(upload_type="RUN", local_path=self.local_path, root_path=root_path, path=path,
                              run_id=run.run_id, artifact_id=0, last_file_yn="Y", teamName=run.team_name,
                              expName=run.exp_name,
                              run_name=run.run_name,
                              artifact_name="", artifact_type="", artifact_digest="", entry_digest=digest,
                              entry_list=[], file_index=0)

    def get_path(self, key_name: str) -> str:
        # server 패스
        return f"{key_name}.IMAGE.{self.format}"

    def get_root_path(self, run_id: int) -> str:
        return os.path.join(str(run_id), "media")



from __future__ import annotations
from deepdriver import logger
experiment: Experiment = None


def set_experiment(experiment_: Experiment) -> None:
    global experiment
    experiment = experiment_


def get_experiment() -> Experiment:
    global experiment
    if not experiment:
        logger.warning("experiment is None")
    return experiment


class Experiment:
    def __init__(self, exp_name, team_name):
        self.__exp_name = exp_name
        self.__team_name = team_name

    @property
    def exp_name(self):
        return self.__exp_name

    @property
    def team_name(self):
        return self.__team_name



import re
from typing import Dict, Callable, Tuple
from urllib.parse import urljoin

import optuna
import ast
from assertpy import assert_that

from optuna.exceptions import DuplicatedStudyError

import deepdriver
from deepdriver import logger
from deepdriver import util
from deepdriver.sdk.config import Config
from deepdriver.sdk.data_types.experiment import set_experiment, Experiment, get_experiment
from deepdriver.sdk.data_types.run import set_run, Run, get_run
from deepdriver.sdk.interface import interface


@util.login_required
def init(exp_name: str = "", team_name: str = "", run_name: str = "", config: Dict = None) -> Run:
    """ # 실행과 실험환경을 만드는 함수 """
    # exp_name 변수 vailidation
    if exp_name:
        pattern = re.compile('[^a-zA-Z0-9._]+')
        if pattern.findall(exp_name):
            logger.info("init() failed : exp_name은 숫자(number), 영문자(alphabet), 언더바(_), 온점(.)만 가능합니다.")
            return None

        if len(exp_name) >= 50:
            logger.info("init() failed :  exp_name의 최대 길이는 50자 미만입니다. [max length 50]")
            return None

    rsp = interface.init(exp_name, team_name, run_name, config)
    run_url = urljoin(f"http://{interface.get_http_host_ip()}:9111", rsp['runUrl'])
    run = Run(rsp["teamName"], rsp["expName"], rsp["runName"], rsp["runId"], run_url)
    logger.info("DeepDriver initialized\n"
                f"Team Name={rsp['teamName']}\n"
                f"Exp Name={rsp['expName']}\n"
                f"Run Name={rsp['runName']}\n"
                f"Run URL={run_url}"
                )
    set_run(run)

    # init config
    deepdriver.config = Config()
    if config:
        for key, value in config.items():
            setattr(deepdriver.config, key, value)

    return run


@util.login_required
def create_hpo(exp_name: str = "", team_name: str = "", remote: bool = False, hpo_config: Dict = None) -> (bool, int):
    # hpo_config['parameters']를 REST API스펙에 맞게 key-value 형식으로 변환
    if hpo_config and 'parameters' in hpo_config:
        parameters_dict = hpo_config['parameters']
        key_value_parameters_list = []
        for key, value in parameters_dict.items():
            key_value_dict = {
                "key": key,
                # "value": {
                #     next(iter(value.keys())): next(iter(value.values()))
                # }
            }
            if 'range' in value.keys():
                key_value_dict['value'] = {'range': value['range']}
                if 'distribution' in value.keys():
                    key_value_dict['description'] = value['distribution']

            if 'values' in value.keys():
                key_value_dict['value'] = {'values': value['values']}

            key_value_parameters_list.append(key_value_dict)

        hpo_config['parameters'] = key_value_parameters_list

    rsp = interface.create_hpo(exp_name, team_name, hpo_config)
    logger.info("HPO initialized\n"
                f"Team Name={rsp['teamName']}\n"
                f"Exp Name={rsp['expName']}\n"
                f"Exp Url={rsp['expUrl']}"
                )
    set_experiment(Experiment(exp_name=rsp['expName'], team_name=rsp['teamName']))
    # optuna 최적화 실행
    if not remote:
        def get_optuna_sampler(hpo_config: dict) -> optuna.samplers.BaseSampler:
            type = hpo_config['method']

            if type.lower() == "random":
                return optuna.samplers.RandomSampler()
            elif type.lower() == "grid":
                search_space = {}
                for item in hpo_config['parameters']:
                    key_ = item['key']
                    value_ = list(item['value'].values())[0]
                    search_space[key_] = value_
                logger.debug(f"search_space : {search_space}")
                return optuna.samplers.GridSampler(search_space)
            elif type.lower() == "bayesian":
                return optuna.samplers.TPESampler()

        team_name_ = team_name or get_experiment().team_name  # team 이름이 입력되지 않은경우 run에 설정된 team 이름을 사용
        try:
            optuna.create_study(study_name=team_name_ + "_" + exp_name,
                                storage="postgresql://hpo:hpo@ce-postg.bokchi.com:5432/hpo",  # TODO : 변경예정
                                direction=hpo_config['metric']['goal'],
                                sampler=get_optuna_sampler(hpo_config),
                                )
        except DuplicatedStudyError:
            logger.info("optuna study aleady exist !")
        except Exception:
            logger.info("please check url ")

    return rsp['result'], rsp['expId']


@util.login_required
def run_hpo(exp_name: str = "", team_name: str = "", remote: bool = False, hpo_config: Dict = None,
            func: Callable = None, count: int = 10, artifact: Dict = None, job_count: int = 0) -> bool:
    # assert_that(func).is_not_none()

    result1, hpoConfig = get_hpo(exp_name=exp_name, team_name=team_name)
    result2, study = load_hpo(exp_name=exp_name, team_name=team_name)

    if artifact:
        if "dataset" in artifact:
            result, dataset_path = prepare_dataset(artifact["dataset"])
            if result:
                hpoConfig['dataset_path'] = dataset_path
        if "code" in artifact:
            result, code_path, func = prepare_code(artifact["code"])
            if result:
                hpoConfig['code_path'] = code_path

    result3 = run_optimize(exp_name=exp_name, team_name=team_name, func=func, config=hpoConfig, count=count,
                           study=study)

    return True


@util.login_required
def prepare_dataset(dataset: dict = None) -> Tuple[bool, str]:
    assert_that(dataset).is_not_none()

    arti = deepdriver.get_artifact(**dataset)
    if arti.entry_list == []:
        logger.error(f"artifat(name={dataset['name']}, type={dataset['type']}) have no entries")
        return False, None
    else:
        path = arti.download()
        return True, path


@util.login_required
def prepare_code(code: dict = None) -> Tuple[bool, str, Callable]:
    assert_that(code).is_not_none()

    arti = deepdriver.get_artifact(name=code['name'], type=code['type'])
    if arti.entry_list == []:
        return False, None, None
    else:
        path = arti.download()
        # finding code file in entry_list
        file_entry_list = next((entry.local_path for entry in arti.entry_list if entry.path.endswith(code['file'])),
                               None)
        if not file_entry_list:
            raise Exception(f'cannot find file({code["file"]}) in artifcat')

        # dynamically function load from file
        with open(file_entry_list, "r") as file:
            file_content = file.read()
        node = ast.parse(file_content)

        looking_for_body = next((body for body in node.body if body.name == code['fun']), None)
        if not looking_for_body:
            raise Exception(f'cannot find function({code["func"]}) in file_entry_list')
        func_name = looking_for_body.name
        func_source = ast.get_source_segment(file_content, looking_for_body)
        exec(func_source)

        return True, path, eval(func_name)


@util.login_required
def get_hpo(exp_name: str = "", team_name: str = "") -> (bool, Dict):
    team_name_ = team_name or get_experiment().team_name
    rsp = interface.get_hpo(exp_name, team_name_)
    return rsp['result'], rsp['hpoConfig']


@util.login_required
def load_hpo(exp_name: str = "", team_name: str = "") -> (bool, optuna.study.study.Study):
    team_name_ = team_name or get_experiment().team_name
    try:
        study = optuna.load_study(study_name=team_name_ + "_" + exp_name,
                                  storage="postgresql://hpo:hpo@ce-postg.bokchi.com:5432/hpo")
    except Exception as e:
        logger.exception(e)
        return False, None
    return True, study


@util.login_required
def run_optimize(exp_name: str = "", team_name: str = "", func: Callable = None, config: Dict = {},
                 count: int = 10, study: optuna.study.study.Study = None) -> bool:
    def get_wrapped_func(exp_name, team_name, func, config):
        return lambda trial: objective(trial, exp_name, team_name, func, config)

    def objective(trial, exp_name, team_name, func, config):
        deepdriver.init(exp_name=exp_name)
        get_config(trial, config)
        y = func()
        deepdriver.log({"result_score": y})
        deepdriver.finish()
        return y

    def get_config(trial, config):
        # config 의 dataset_path, code_path를 config에 저장
        if ("dataset_path" in config) and ("code_path" in config):
            deepdriver.config.dataset_path = config['dataset_path']
            deepdriver.config.code_path = config['code_path']

        # config 의 parameters 내용을 조건에  따라 하기와 같은 코드로 변환
        for item_dict in config['parameters']:
            key: str = item_dict['key']
            sub_key = next(iter(item_dict['value'].keys()))  # 'values' or 'range'
            values: list = next(iter(item_dict['value'].values()))
            suggest = None
            if sub_key == "range":
                if 'description' in item_dict and item_dict['description']:
                    # 'description'(distribution)이 있는 경우
                    if item_dict['description'] == "uniform":
                        suggest = trial.suggest_uniform(item_dict['description'], min(values), max(values))
                    if item_dict['description'] == "log_uniform":
                        suggest = trial.suggest_loguniform(item_dict['description'], min(values), max(values))
                else:
                    if all(isinstance(x, float) for x in values):
                        suggest = trial.suggest_float(key, min(values), max(values))
                    if all(isinstance(x, int) for x in values):
                        suggest = trial.suggest_int(key, min(values), max(values))
                if suggest:
                    exec(f'deepdriver.config.{key} = suggest')
                else:
                    logger.error(f"error : {item_dict}")
            elif sub_key == "values":
                exec(f'deepdriver.config.{key} = trial.suggest_categorical("{key}", {values})')

        logger.debug(f"deepdriver.config. : {deepdriver.config.__dict__}")

    team_name_ = team_name or get_experiment().team_name
    try:
        study = study.optimize(get_wrapped_func(exp_name, team_name_, func, config), n_trials=count)
    except Exception as e:
        logger.exception(e)
        return False, None
    return True, study



import logging
import sys
from deepdriver.sdk import util
import os

# Google introduced an incompatibility into protobuf-4.21.x
# that is not backwards compatible with many libraries.
# Once those libraries have updated to rebuild their _pb2.py files,
# this can be removed.
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

try:
  from google.protobuf.internal import builder as _builder
except ImportError:
  # tensorflow와 protobuf 버젼 호환 문제로 protovuf 4.2.x 버젼을 사용하게 함
  raise ImportError("A higher protobuf package version is requried. Upgrade protobuf using 'pip install --upgrade protobuf==4.21.12'" )

logger = logging.getLogger("deepdriver")
logger.propagate = False
if logger.handlers == []:
  console_handler = logging.StreamHandler(sys.stdout)
  if not util.is_notebook():
    formatter = logging.Formatter('%(asctime)s %(levelname)s [%(name)s] [%(filename)s:%(lineno)d] - %(message)s')
    console_handler.setFormatter(formatter)
  else:
    formatter = logging.Formatter('%(message)s')
    console_handler.setFormatter(formatter)
  logger.setLevel(logging.INFO)
  logger.addHandler(console_handler)


from deepdriver.sdk.artifact import *
from deepdriver.sdk.config import *
from deepdriver.sdk.experiment import *
from deepdriver.sdk.login import *
from deepdriver.sdk.setting import *
from deepdriver.sdk.run import *
from deepdriver.sdk.visualization import visualize
from deepdriver.sdk.chart.histogram import histogram
from deepdriver.sdk.chart.line import line
from deepdriver.sdk.chart.scatter import scatter
from deepdriver.sdk.chart.confusion_matrix import confusion_matrix
from deepdriver.sdk.chart.roc_curve import roc_curve
from deepdriver.sdk.data_types.dataFrame import DataFrame
from deepdriver.sdk.data_types.image import Image
from deepdriver.sdk.data_types.table import Table


from deepdriver.sdk.lib.lazyloader import LazyLoader
keras = LazyLoader('deepdriver.intergration.keras', globals(), 'deepdriver.intergration.keras.keras')  #layzyload

config = Config()




from typing import Dict

import tensorflow as tf
import copy

from deepdriver.sdk.data_types.run import get_run
from deepdriver import logger


class MLCallback(tf.keras.callbacks.Callback):

    def __init__(self):
        pass

    def on_epoch_end(self, epoch: int, logs: Dict = {}):
        logger.debug(f"logs:{logs}")
        logger.debug(f"epoch:{epoch}")

        to_send_log = copy.deepcopy(logs)
        if "accuracy" in logs:
            to_send_log["acc"] = logs["accuracy"]

        if "val_accuracy" in logs:
            to_send_log["val_acc"] = logs["val_accuracy"]

        logger.debug(f"to_send_logs:{to_send_log}")
        get_run().log(to_send_log)



from assertpy import assert_that
import requests
from deepdriver.sdk.interface import interface
from deepdriver import logger
# deepdriver 실험환경을 사용하기위한 로그인 과정
# 서버의 login api를 호출하여 key를 서버로 전송하고 결과로서 jwt key를 받는다
def login(key: str) -> bool:
    assert_that(key).is_not_none()

    try:
        return interface.login(key)
    except (requests.exceptions.ConnectTimeout, requests.exceptions.ConnectionError) as e:
        logger.error(f"Could Login to Server[{interface.get_http_host()}]. Set Server IP/PORT using deepdriver.setting()")
        return False






import numpy as np
from assertpy import assert_that

from deepdriver.sdk.chart.chart import Chart, TYPE_CONFUSION_MATRIX


def confusion_matrix(y_true: list, probs: list = None, preds: list = None, class_names: list = None,
                     title: str = None) -> Chart:
    assert_that(y_true).is_not_none()
    if (probs is None) and (preds is None):
        raise Exception("probs, preds one must have value")
        return

    if (probs is not None) and (preds is None):
        preds = np.argmax(probs, axis=1).tolist()

    if class_names is None:
        class_inds = set(preds).union(set(y_true))
        n_classes = len(class_inds)
        class_names = [f"Class_{i}" for i in range(1, n_classes + 1)]
    else:
        n_classes = len(class_names)
        class_inds = [i for i in range(n_classes)]

    class_mapping = {}
    for i, val in enumerate(sorted(list(class_inds))):
        class_mapping[val] = i
    counts = np.zeros((n_classes, n_classes))
    for i in range(len(preds)):
        counts[class_mapping[y_true[i]], class_mapping[preds[i]]] += 1

    data = []
    for i in range(n_classes):
        for j in range(n_classes):
            data.append([class_names[i], class_names[j], counts[i, j]])

    from deepdriver import DataFrame
    dataframe = DataFrame(columns=["Actual", "Predicted", "nPredictions"], data=data)

    fields = {
        "Actual": "Actual",
        "Predicted": "Predicted",
        "nPredictions": "nPredictions",
    }
    return Chart(chart_type=TYPE_CONFUSION_MATRIX, data=dataframe, data_fields=fields, label_fields={"title": title})



import hashlib
import os
import sys
from pathlib import Path
from typing import Dict, List, Union

from assertpy import assert_that

from deepdriver.sdk.chart.chart import Chart
from deepdriver.sdk.data_types.image import Image
from deepdriver.sdk.data_types.run import get_run
from deepdriver.sdk.data_types.table import Table
from deepdriver.sdk.interface import interface
from deepdriver.sdk.interface.grpc_interface_pb2 import ArtifactEntry as grpc_ArtifactEntry


class ArtifactEntry:

    def __init__(self, path: str, local_path: str, size: int, digest: str, status: str, lfs_yn: str, repo_tag: str,
                 type: str, metadata: str, key: str = ""):
        assert_that(path).is_not_none()
        assert_that(local_path).is_not_none()

        self.path = path  # path는 서버에 전송될 경로
        self.local_path = local_path  # local_path는 artifact.add()를 통해서 추가된 파일의 로컬 경로
        self.size = size  # 파일의 사이즈
        self.digest = digest  # 파일의 digest 값
        self.status = status  # 파일의 상태(로컬과의 비교)
        self.lfs_yn = lfs_yn
        self.repo_tag = repo_tag
        self.type = type
        self.metadata = metadata
        self.key = key

    # ArtifactEntry 객체를 기반으로 파일 다운로드 수행
    def download(self, local_root_path: str, artifact_id: int, team_name: str, exp_name: str, artifact_name: str,
                 artifact_type: str, versioning: str, file_index: int, total_file_count: int):
        assert_that(local_root_path).is_not_none()
        assert_that(artifact_id).is_not_none()

        local_path = os.path.join(local_root_path, self.repo_tag, self.path)
        if sys.platform.startswith('win32'):
            # 윈도우인 경우 Path Seperator를 리눅스 형식으로 강제로 변경함
            local_path = local_path.replace("\\", "/")
        i_last_separator = len(local_path) - local_path[::-1].find("/") - 1
        Path(local_path[:i_last_separator]).mkdir(parents=True, exist_ok=True)

        interface.download_file(self.path, artifact_id, local_path, team_name, exp_name, artifact_name, artifact_type,
                                versioning, self.lfs_yn, self.repo_tag, file_index, total_file_count, self.size)
        # 다운로드 후 파일의 digest 확인
        with open(local_path, "rb") as f:
            downloaded_file_digest = hashlib.md5(f.read()).hexdigest()
        if self.digest != downloaded_file_digest:
            raise Exception(
                f"file check sum error : checksum value are '{self.digest}' expected, but '{downloaded_file_digest}'")

        # 다운로드 완료 후 local_path 설정
        self.local_path = local_path

    # ArtifactEntry 객체를 기반으로 파일 업로드 수행
    def upload(self, upload_type: str, root_path: str, run_id: int, artifact_id: int, last_file_yn: str, team_name: str,
               exp_name: str, run_name: str, artifact_name: str, artifact_type: str, artifact_digest: str,
               entry_list: List[grpc_ArtifactEntry], file_index: int) -> bool:
        assert_that(root_path).is_not_none()
        assert_that(run_id).is_not_none()
        assert_that(artifact_id).is_not_none()
        assert_that(last_file_yn).is_not_none()
        assert_that(artifact_digest).is_not_none()
        assert_that(entry_list).is_not_none()
        if self.local_path:
            return interface.upload_file(upload_type, self.local_path, root_path, self.path, run_id, artifact_id,
                                         last_file_yn, team_name, exp_name, run_name, artifact_name, artifact_type,
                                         artifact_digest, self.digest, entry_list, file_index=file_index)
        else:
            return True

    def __str__(self) -> str:
        return str(self.__dict__)


# Artifact 객체에 추가되는 객체는 Image, 일반 파일, 폴더(하위 파일 포함), Table, 소스코드 등이 있다
class Artifacts:

    def __init__(self, name: str, type: str, id: int = 0, desc: str = None, versioning: str = "Y",
                 meta_data: Dict = None, entry_list: List[ArtifactEntry] = None) -> None:
        assert_that(name).is_not_none()
        assert_that(type).is_not_none()

        self.name = name
        self.type = type
        self.id = id
        self.desc = desc
        self.versioning = versioning
        self.run = get_run()

        # 메타정보 (for 모델 or 코드)
        # Ai 학습 모델과 같은 데이터가 아티펙트로 추가될 때 추가정보( 하이퍼파라미터)도 같이 기록되기위한 meta_data 정보
        self.meta_data: Dict = meta_data if meta_data else {}

        # ArtifactEntry 의 list
        # Artifacts.add() 함수로 데이터가 추가될때마다 리스트에 추가됨
        self.entry_list: List[ArtifactEntry] = entry_list if entry_list else []

    # 해당 Artifacts의 파일을 다운로드 후에 key에 해당하는 객체를 찾아 객체로 복원해서 반환
    def get(self, key: str = None) -> Union[str, Table, Image, Chart]:
        self.download()
        try:
            artifact_entry = next(item for item in self.entry_list if item.key == key)
        except StopIteration as e:
            raise Exception(f"Cannnot find Entry by Key: key=[{key}]")

        if artifact_entry.type == "TABLE":
            return Table.from_file(artifact_entry.local_path)
        if artifact_entry.type == "IMAGE":
            return Image(artifact_entry.local_path)
        if artifact_entry.type == "CHART":
            return Chart.from_file(artifact_entry.local_path)

    # 해당 Artifact에 데이터를 추가하는 함수
    # obj 및 파일 내용이 추가되면 임시 파일로 저장후 path를 추가한다
    # 추가된 각 항목에 대해서 ArtifactEntry를 생성한 후 Artifact가 가진 ArtifactEntry list에 추가함
    # name: 아티펙트에 각 엔트리의 식별자로 기록될 이름
    # data: 아티펙트에 추가될 데이터( Table  | Image | chart | file name | directory path | reference url )
    def add(self, data: Union[str, Table, Image, Chart], name: str = None) -> None:
        if isinstance(data, str):
            # if data is file_path
            data = os.path.relpath(data)  # Convert any path to a relative path
            data = os.path.join(".", data)  # Prefix with a dot
            self.__add_file(data, name)
            # local 파일의 존재여부를 판단하여 "DELETE" 상태로 마킹
            self.__local_path_sync_for_delete()
        elif isinstance(data, Table):
            # if data is Table
            assert_that(name).is_not_none()
            # TODO: local_path?
            local_path = os.path.join(".", data.get_path(name))
            digest, size = data.file_dump(local_path)
            self.entry_list.append(
                ArtifactEntry(path=data.get_path(name), local_path=local_path, size=size, digest=digest, status="ADD",
                              lfs_yn="", repo_tag="", type="TABLE", metadata=data.to_json(name), key=name))
        elif isinstance(data, Image):
            assert_that(name).is_not_none()
            local_path = data.local_path
            digest = data.get_digest()
            size = data.get_size()
            self.entry_list.append(
                ArtifactEntry(path=data.get_path(name), local_path=local_path, size=size, digest=digest, status="ADD",
                              lfs_yn="", repo_tag="", type="IMAGE", metadata=data.to_json(name), key=name))
        elif isinstance(data, Chart):
            assert_that(name).is_not_none()
            # TODO: local_path?
            local_path = os.path.join(".", data.get_path(name))
            digest, size = data.file_dump(local_path)
            self.entry_list.append(
                ArtifactEntry(path=data.get_path(name), local_path=local_path, size=size, digest=digest, status="ADD",
                              lfs_yn="", repo_tag="", type="CHART", metadata=data.to_json(name), key=name))

    def __add_file(self, data: Union[str], name: str = None, depth: int = 0) -> None:
        # data에 str가 들어오면 file 인지, dir 인지 reference url(http:// or s3:// ..) 인지 체크
        if isinstance(data, str):
            if not os.path.isfile(data):
                # dir인경우 하위 파일을 순회하며 위의 과정 반복
                dir_path = data
                for file_dir_name in os.listdir(dir_path):
                    file_dir_path = os.path.join(dir_path, file_dir_name)
                    # add시 name이 지정된 경우: name/하위폴더명/파일명
                    # add시 name이 지정되지 않은 경우: 하위폴더명/파일명
                    target_name = file_dir_name if depth == 0 and not name else name
                    self.__add_file(file_dir_path, target_name, depth + 1)
            else:
                # file인경우 ArtifactEntry 생성후 local_path, path,size, digest등을 설정한 후 Artifact의 entry_list에 추가
                local_path = data
                fd_names = local_path.split(os.path.sep)
                for i, fd_name in enumerate(fd_names):
                    if fd_name == name:
                        fd_names = fd_names[i:]
                        break
                path = os.path.join(*fd_names)
                if sys.platform.startswith('win32'):
                    # 윈도우인 경우 Path Seperator를 리눅스 형식으로 강제로 변경함
                    path = path.replace("\\", "/")
                size = os.stat(local_path).st_size
                with open(local_path, "rb") as f:
                    local_file_digest = hashlib.md5(f.read()).hexdigest()

                entry = self.__find_entry_by_path(local_path, path)
                if entry:
                    entry.status = "SYNC"  # 변경이 없는 경우
                    if entry.digest != local_file_digest:   # local의 root 경로가 바껴서 파일명은 동일하지만, 파일내용이 바뀐경우
                        entry.local_path = local_path
                        entry.digest = local_file_digest
                        entry.size = size
                else:
                    self.entry_list.append(
                        ArtifactEntry(path, local_path, size, local_file_digest, status="ADD", lfs_yn="", repo_tag="", type="FILE",
                                      metadata=""))  # 파일이 추가된 경우

    def __local_path_sync_for_delete(self) -> None:
        # local에서 삭제된 파일의 status를 "DELETE"로 변경
        for entry in self.entry_list:
            if entry.status == "ADD":  # 상태가 "ADD"인 entry만 확인
                if not os.path.isfile(entry.local_path):
                    entry.status = "DELETE"

    def __find_entry_by_path(self, local_path: str, path: str) -> ArtifactEntry:
        for entry in self.entry_list:
            if not entry.local_path:  # download()를 실행하지 않은 경우 local_path가 ""이므로 entry.path에서 비교
                if entry.path == path:
                    return entry

            if entry.local_path == local_path:
                return entry
        else:
            return None

    # Artifact의 정보를 전송하고, ArtifactEntry 각각의 파일을 전송한다
    def upload(self) -> bool:
        # Interface.py의 upoad_artifact ( grpc_interface.py의 upoad_artifact)를 호출 하여 artifact 정보에 대해 전송
        entry_dict = {entry.path: entry for entry in self.entry_list}
        entry_list: List(grpc_ArtifactEntry) = []
        hasher = hashlib.md5()
        # TODO: 상태에 따라서 status 변경
        for entry_path, entry in sorted(entry_dict.items()):
            entry_list.append(grpc_ArtifactEntry(
                path=entry_path,
                digest=entry.digest,
                size=entry.size,
                status="ADD",
                key=entry.key,
                type=entry.type,
                metadata=entry.metadata,
            ))
            hasher.update(f"{entry_path}:{entry.digest}\n".encode())
        artifact_digest = hasher.hexdigest()
        id = interface.upload_artifact(self.run.run_id, self, artifact_digest, self.run.team_name, self.run.exp_name,
                                       entry_list)
        if id is None:
            return False
        self.id = id

        # artifact에 추가된 파일들(ArtifactEntry의 파일)은 각각의 upload함수 호출
        for i, entry in enumerate(self.entry_list):
            # ArtifactEntry.upload() 호출시 인자로서 파일이 저장될 폴더 정보를 넘겨줌
            root_path = os.path.join(str(self.run.run_id), "artifact", self.type, self.name)
            last_file_yn = "Y" if i == len(self.entry_list) - 1 else "N"
            if self.versioning == "Y":
                upload_type = "ARTI_REPO"
            else:
                upload_type = "ARTI_FILES"

            succeeded = entry.upload(upload_type, root_path, self.run.run_id, self.id, last_file_yn, self.run.team_name,
                                     self.run.exp_name, self.run.run_name, self.name, self.type, artifact_digest,
                                     entry_list, file_index=i)
            if not succeeded:
                return False
        return True

    # Artifact 객체를 기반으로 Artifact에 등록된 모든 엔트리의 다운로드 수행
    def download(self) -> str:
        local_root_path = self.get_download_dir(with_repo_tag=False)
        # Artifact 객체가 가진 entry_list 의  ArtifactEntry.download()를 각각 호출
        for idx, entry in enumerate(self.entry_list):
            entry.download(local_root_path, self.id, self.run.team_name, self.run.exp_name, self.name, self.type,
                           self.versioning, idx, total_file_count=len(self.entry_list))
        return local_root_path

    def get_download_dir(self, with_repo_tag: bool = True):
        # ./deepdriver/artifact/{artifact_id}/{tag} 폴더가 생성된다
        download_path = os.path.join(".", "deepdriver", "artifact", str(self.id))

        if with_repo_tag and len(self.entry_list) > 0:
            return os.path.join(download_path, self.entry_list[0].repo_tag)
        else:
            return download_path

    def __str__(self) -> str:
        return "[" + ",".join(str(entry) for entry in self.entry_list) + "]"



import json
import pandas as pd
from assertpy import assert_that
from typing import List

class DataFrame:

    def __init__(self, data: List[List]=None, columns: List[str]=None, dataframe: pd.DataFrame=None) -> None:
        # columns와 data, dataframe중 하나의 파라미터는 입력이 되어야함
        if dataframe is not None:
            # dataframe이 입력된 경우 멤버변수의 dataframe으로 설정하고
            self.dataframe = dataframe

            # columns와 data 멤버변수를 dataframe으로부터 추출하여 설정한다
            self.data: List[List] = dataframe.values.tolist()
            self.columns: List[str] = dataframe.columns.tolist()
        else:
            assert_that(data).is_not_none()
            assert_that(columns).is_not_none()

            # columns와 data가 입력된 경우 추후 visualize에서 사용하기 쉽도록 pd.DataFrame를 columns와 data를 통해 생성하고 멤버변수 dataframe에 설정한다
            self.dataframe = pd.DataFrame(data=data, columns=columns)
            self.data = data
            self.columns = columns

    def to_dict(self) -> str:
        return {
            "columns" : self.columns,
            "data" : self.data,
        }
